import json
import os
from datetime import datetime

import numpy as np
import pandas as pd

from pyscnomics.contracts.project import BaseProject
from pyscnomics.contracts.costrecovery import CostRecovery
from pyscnomics.contracts.grossplit import GrossSplit
from pyscnomics.econ.costs import CapitalCost, Intangible, OPEX, ASR
from pyscnomics.econ.revenue import Lifting, FluidType


def get_json_file_names() -> list:
    """
    A function to get a list of json file in the chosen directory.

    Returns
    -------
    list_json_file
        The of available dataset.
    """

    list_json_file = []
    if 'pyscnomics\\pyscnomics' in os.path.dirname(os.getcwd()):
        directory = (os.path.dirname(os.getcwd())) + '/dataset'

    else:
        directory = (os.path.dirname(os.getcwd())) + '/pyscnomics/dataset'

    for filenames in os.listdir(os.path.abspath(directory)):
        if filenames.endswith(".json"):
            full_file_name = filenames
            file_name = os.path.splitext(full_file_name)[0]
            list_json_file.append(file_name)

    return list_json_file


def read_json_file(file_name: str) -> dict:
    """
    A function to read json file.

    Parameters
    ----------
    file_name: str
        The file name of the json file.

    Returns
    -------
    json.loads(file_contents)
        The dictionary of the json file.

    """
    if 'pyscnomics\\pyscnomics' in os.path.dirname(os.getcwd()):
        directory = (os.path.dirname(os.getcwd())) + '/dataset/' + file_name + '.json'

    else:
        directory = (os.path.dirname(os.getcwd())) + '/pyscnomics/dataset/' + file_name + '.json'

    final_dir = os.path.abspath(directory)
    with open(final_dir) as user_file:
        file_contents = user_file.read()

    return json.loads(file_contents)


def read_fluid_type(fluid: list | str) -> list[FluidType] | FluidType:
    """
    A function to converting the str input into FluidType Enum class.

    Parameters
    ----------
    fluid: str | list
        The fluid type.

    Returns
    -------
    list[FluidType] | FluidType
        The enum class for FluidType.

    """
    if isinstance(fluid, str):
        if fluid == 'Oil':
            return FluidType.OIL
        elif fluid == 'Gas':
            return FluidType.GAS
        elif fluid == 'Sulfur':
            return FluidType.SULFUR
        elif fluid == 'Electricity':
            return FluidType.ELECTRICITY
        elif fluid == 'CO2':
            return FluidType.CO2

    else:

        fluid_mapping = {'Oil': FluidType.OIL,
                         'Gas': FluidType.GAS,
                         'Sulfur': FluidType.SULFUR,
                         'Electricity': FluidType.ELECTRICITY,
                         'CO2': FluidType.CO2}

        # Replace elements in the list using the mapping
        result = [fluid_mapping[i] for i in fluid if i in fluid_mapping]

        return result


def assign_onstream_date(date_data: None | str) -> datetime | None:
    """
    A funtion to assign onstream date into datetime format.
    Parameters
    ----------
    date_data: None | str
        date data read from the json file or Ms. Excel file.

    Returns
    -------
    result_date: datetime
        The date in the datetime format

    """
    if date_data is not None:
        result_date = datetime.strptime(date_data, '%d/%m/%Y').date()
    else:
        result_date = None

    return result_date


def assign_lifting(data_raw: dict) -> tuple:
    """
    A function to assign the lifting data to the list of Lifting dataclass.

    Parameters
    ----------
    data_raw: dict
        The dictionary containing the lifting data.

    Returns
    -------
    lifting_list: tuple
        The list containing the lifting dataclass.

    """
    # Defining the data source and the list container for Lifting. Then, assign them based on their fluid type
    lifting_data = data_raw['lifting']
    lifting_list = []
    for key in lifting_data.keys():
        # Since the Lifting data for gas has different arguments input, conditional formatting is applied
        if 'Gas' in key or 'GSA' in key:
            lifting = Lifting(start_year=lifting_data[key]["start_year"],
                              end_year=lifting_data[key]["end_year"],
                              lifting_rate=np.array(lifting_data[key]["lifting_rate"]),
                              price=np.array(lifting_data[key]["price"]),
                              prod_year=np.array(lifting_data[key]["prod_year"]),
                              fluid_type=read_fluid_type(lifting_data[key]["fluid_type"]),
                              ghv=None if lifting_data[key]["ghv"] is None else
                              np.array(lifting_data[key]["ghv"]),
                              prod_rate=None if lifting_data[key]["prod_rate"] is None
                              else np.array(lifting_data[key]["prod_rate"]),
                              # prod_rate_baseline=None if lifting_data[key]["prod_rate_baseline"] is None
                              # else np.array(lifting_data[key]["prod_rate_baseline"]),
                              )

            lifting_list.append(lifting)

        else:
            lifting = Lifting(start_year=lifting_data[key]["start_year"],
                              end_year=lifting_data[key]["end_year"],
                              lifting_rate=np.array(lifting_data[key]["lifting_rate"]),
                              price=np.array(lifting_data[key]["price"]),
                              prod_year=np.array(lifting_data[key]["prod_year"]),
                              fluid_type=read_fluid_type(lifting_data[key]["fluid_type"]),
                              prod_rate=None if lifting_data[key]["prod_rate"] is None
                              else np.array(lifting_data[key]["prod_rate"]),
                              prod_rate_baseline=None if lifting_data[key]["prod_rate_baseline"] is None
                              else np.array(lifting_data[key]["prod_rate_baseline"]),
                              )
            lifting_list.append(lifting)

    return tuple(lifting_list)


def assign_cost(data_raw: dict) -> tuple:
    """
    A function to assign the cost data to the list of Tangible, Intangible, OPEX and ASR dataclass.

    Parameters
    ----------
    data_raw: dict

    Returns
    -------
    tangible_list : tuple
        A tuple containing the Tangible dataclass.
    intangible_list : tuple
        A tuple containing the Intangible dataclass.
    opex_list
        A tuple containing the OPEX dataclass.
    asr_list
        A tuple containing the ASR dataclass.

    """
    # Defining the data source and the list container for Tangible. Then, assign them based on their fluid type
    tangible_data = data_raw['tangible']
    tangible_list = []
    for key in tangible_data.keys():
        tangible = CapitalCost(start_year=tangible_data[key]['start_year'],
                               end_year=tangible_data[key]['end_year'],
                               cost=np.array(tangible_data[key]['cost']),
                               expense_year=np.array(tangible_data[key]['expense_year']),
                               cost_allocation=read_fluid_type(fluid=tangible_data[key]['cost_allocation']),
                               description=tangible_data[key]['description'],
                               vat_portion=np.array(tangible_data[key]['vat_portion']),
                               vat_discount=np.array(tangible_data[key]['vat_discount']),
                               lbt_portion=np.array(tangible_data[key]['lbt_portion']),
                               lbt_discount=np.array(tangible_data[key]['lbt_discount']),
                               pis_year=np.array(tangible_data[key]['pis_year']),
                               salvage_value=np.array(tangible_data[key]['salvage_value']),
                               useful_life=np.array(tangible_data[key]['useful_life']),
                               depreciation_factor=np.array(tangible_data[key]['depreciation_factor']),
                               is_ic_applied=tangible_data[key]['is_ic_applied'],
                               )
        tangible_list.append(tangible)

    # Defining the data source and the list container for Intangible. Then, assign them based on their fluid type
    intangible_data = data_raw['intangible']
    intangible_list = []
    for key in intangible_data.keys():
        intangible = Intangible(start_year=intangible_data[key]['start_year'],
                                end_year=intangible_data[key]['end_year'],
                                cost=np.array(intangible_data[key]['cost']),
                                expense_year=np.array(intangible_data[key]['expense_year']),
                                cost_allocation=read_fluid_type(fluid=intangible_data[key]['cost_allocation']),
                                description=intangible_data[key]['description'],
                                vat_portion=np.array(intangible_data[key]['vat_portion']),
                                vat_discount=np.array(intangible_data[key]['vat_discount']),
                                lbt_portion=np.array(intangible_data[key]['lbt_portion']),
                                lbt_discount=np.array(intangible_data[key]['lbt_discount']),
                                )
        intangible_list.append(intangible)

    # Defining the data source and the list container for Intangible. Then, them based on their fluid type
    opex_data = data_raw['opex']
    opex_list = []
    for key in opex_data.keys():
        opex = OPEX(start_year=opex_data[key]['start_year'],
                    end_year=opex_data[key]['end_year'],
                    expense_year=np.array(opex_data[key]['expense_year']),
                    cost_allocation=read_fluid_type(fluid=opex_data[key]['cost_allocation']),
                    description=opex_data[key]['description'],
                    vat_portion=np.array(opex_data[key]['vat_portion']),
                    vat_discount=np.array(opex_data[key]['vat_discount']),
                    lbt_portion=np.array(opex_data[key]['lbt_portion']),
                    lbt_discount=np.array(opex_data[key]['lbt_discount']),
                    fixed_cost=np.array(opex_data[key]['fixed_cost']))
        opex_list.append(opex)

    # Defining the data source and the list container for ASR. Then, them based on their fluid type
    asr_data = data_raw['asr']
    asr_list = []
    for key in asr_data.keys():
        asr = ASR(start_year=asr_data[key]['start_year'],
                  end_year=asr_data[key]['end_year'],
                  cost=np.array(asr_data[key]['cost']),
                  expense_year=np.array(asr_data[key]['expense_year']),
                  cost_allocation=read_fluid_type(fluid=asr_data[key]['cost_allocation']),
                  description=asr_data[key]['description'],
                  vat_portion=np.array(asr_data[key]['vat_portion']))
        asr_list.append(asr)

    return tangible_list, intangible_list, opex_list, asr_list


def load_testing(dataset_type: str, key: str) -> dict | ValueError:
    """
    A function to load the available test dataset.

    Parameters
    ----------
    dataset_type
    key

    Returns
    -------

    """
    data = read_json_file(file_name=dataset_type)
    # Condition where the testing data is not available within the chosen dataset
    if data.get("testing") is None:
        return ValueError("The testing data for this dataset is still in development.")

    else:
        data_test = dict(data)['testing'][key]
        return data_test


def load_data(dataset_type: str, contract_type: str = 'project') -> BaseProject | CostRecovery | GrossSplit:
    """
    A function to load the available dataset.

    Parameters
    ----------
    dataset_type: str
        The type of the dataset.
    contract_type
        The type of the contract.
    Returns
    -------
    BaseProject | CostRecovery | GrossSplit

    """
    # Checking the availability of the dataset, if not available raise a Value Error
    dataset_list = get_json_file_names()
    contract_list = ['project', 'cost_recovery', 'gross_split']
    if dataset_type not in dataset_list or contract_type not in contract_list:
        raise ValueError('Unknown dataset: "{0}", please check the Dataset and Contract Type that available.'
                         .format(dataset_type))

    # Reading the project json file
    data_raw = read_json_file(file_name=dataset_type)

    # Assigning the general config and converting them to datetime format.
    project_start_date = datetime.strptime(data_raw['start_date'], '%d/%m/%Y').date()
    project_end_date = datetime.strptime(data_raw['end_date'], '%d/%m/%Y').date()

    oil_onstream_date = assign_onstream_date(data_raw['oil_onstream_date'])
    gas_onstream_date = assign_onstream_date(data_raw['gas_onstream_date'])

    # Assigning the lifting and cost data.
    lifting_list = assign_lifting(data_raw)
    tangible_list, intangible_list, opex_list, asr_list = assign_cost(data_raw)

    if contract_type == 'project':
        return BaseProject(start_date=project_start_date,
                           end_date=project_end_date,
                           oil_onstream_date=oil_onstream_date,
                           gas_onstream_date=gas_onstream_date,
                           lifting=lifting_list,
                           capital_cost=tangible_list,
                           intangible_cost=intangible_list,
                           opex=opex_list,
                           asr_cost=asr_list)

    elif contract_type == 'cost_recovery':
        config = read_json_file(file_name=contract_type)
        return CostRecovery(start_date=project_start_date,
                            end_date=project_end_date,
                            oil_onstream_date=oil_onstream_date,
                            gas_onstream_date=gas_onstream_date,
                            lifting=lifting_list,
                            capital_cost=tangible_list,
                            intangible_cost=intangible_list,
                            opex=opex_list,
                            asr_cost=asr_list,
                            oil_ftp_is_available=config['oil_ftp_is_available'],
                            oil_ftp_is_shared=config['oil_ftp_is_shared'],
                            oil_ftp_portion=config['oil_ftp_portion'],
                            gas_ftp_is_available=config['gas_ftp_is_available'],
                            gas_ftp_is_shared=config['gas_ftp_is_shared'],
                            gas_ftp_portion=config['gas_ftp_portion'],
                            oil_ctr_pretax_share=config['oil_ctr_pretax_share'],
                            gas_ctr_pretax_share=config['gas_ctr_pretax_share'],
                            oil_ic_rate=config['oil_ic_rate'],
                            gas_ic_rate=config['gas_ic_rate'],
                            ic_is_available=config['ic_is_available'],
                            oil_cr_cap_rate=config['oil_cr_cap_rate'],
                            gas_cr_cap_rate=config['gas_cr_cap_rate'],
                            oil_dmo_volume_portion=config['oil_dmo_volume_portion'],
                            oil_dmo_fee_portion=config['oil_dmo_fee_portion'],
                            oil_dmo_holiday_duration=config['oil_dmo_holiday_duration'],
                            gas_dmo_volume_portion=config['gas_dmo_volume_portion'],
                            gas_dmo_fee_portion=config['gas_dmo_fee_portion'],
                            gas_dmo_holiday_duration=config['gas_dmo_holiday_duration'])

    elif contract_type == 'gross_split':
        config = read_json_file(file_name=contract_type)
        return GrossSplit(start_date=project_start_date,
                          end_date=project_end_date,
                          oil_onstream_date=oil_onstream_date,
                          gas_onstream_date=gas_onstream_date,
                          lifting=lifting_list,
                          capital_cost=tangible_list,
                          intangible_cost=intangible_list,
                          opex=opex_list,
                          asr_cost=asr_list,
                          field_status=config['field_status'],
                          field_loc=config['field_loc'],
                          res_depth=config['res_depth'],
                          infra_avail=config['infra_avail'],
                          res_type=config['res_type'],
                          api_oil=config['api_oil'],
                          domestic_use=config['domestic_use'],
                          prod_stage=config['prod_stage'],
                          co2_content=config['co2_content'],
                          h2s_content=config['h2s_content'],
                          base_split_ctr_oil=config['base_split_ctr_oil'],
                          base_split_ctr_gas=config['base_split_ctr_gas'],
                          split_ministry_disc=config['split_ministry_disc'],
                          oil_dmo_volume_portion=config['oil_dmo_volume_portion'],
                          oil_dmo_fee_portion=config['oil_dmo_fee_portion'],
                          oil_dmo_holiday_duration=config['oil_dmo_holiday_duration'],
                          gas_dmo_volume_portion=config['gas_dmo_volume_portion'],
                          gas_dmo_fee_portion=config['gas_dmo_fee_portion'],
                          gas_dmo_holiday_duration=config['gas_dmo_holiday_duration'])


def load_cost(filename: str,
              start_year: int = 2023,
              end_year: int = 2043,
              cost_allocation: FluidType = FluidType.OIL,
              template: str = "pyscnomics") -> tuple[CapitalCost, Intangible, OPEX, ASR] | ValueError:
    """
    Function to load the cost data from Excel file.

    Parameters
    ----------
    filename: str
        The name of the Excel file.
    start_year: int
        The start year of the cost data.
    end_year: int
        The end year of the cost data
    cost_allocation: FluidType
        The fluid type of that the cost will be allocated to.
    template: str
        The type of Excel source that will be read. The available types are: ['pyscnomics', 'questor']

    Returns
    -------
    out: tuple
        Tangible
            The Tangible dataclass.
        Intangible
            The Intangible dataclass
        OPEX
            The OPEX dataclass
        ASR
            The ASR dataclass
    """

    # Defining the available template list and making the condition if not satisfied
    template_list = ['pyscnomics', 'questor']
    if template not in template_list:
        raise ValueError('Unknown Template: "{0}", please check the Template Type in Docstring.'.format(template))

    # Reading the Questor Excels file from column B to W and replacing the value of NaN with 0
    df = pd.read_excel(filename, skiprows=18, header=None, na_values=0).fillna(value=0)
    years_arr = np.arange(start_year, start_year + df.shape[0], 1)
    df.set_index(years_arr, inplace=True)

    # Assigning the Tangible data
    tangible_arr = np.array(df[[5, 7, 8, 9, 10, 11, 12]].sum(axis=1).to_numpy(dtype=float))
    tangible = CapitalCost(start_year=start_year,
                           end_year=end_year,
                           cost=tangible_arr,
                           expense_year=years_arr,
                           cost_allocation=cost_allocation)

    # Assigning the Intangible data
    intangible_arr = df[6].to_numpy(dtype=float)
    intangible = Intangible(start_year=start_year,
                            end_year=end_year,
                            cost=intangible_arr,
                            expense_year=years_arr,
                            cost_allocation=cost_allocation)

    # Assigning the ASR data
    asr_arr = df[18].to_numpy(dtype=float)
    asr = ASR(start_year=start_year,
              end_year=end_year,
              cost=asr_arr,
              expense_year=years_arr,
              cost_allocation=cost_allocation)

    # Assigning the OPEX data
    fixed_cost_arr = df[[13, 14, 15, 16, 17]].sum(axis=1).to_numpy(dtype=float)
    opex = OPEX(start_year=start_year,
                end_year=end_year,
                fixed_cost=fixed_cost_arr,
                expense_year=years_arr,
                cost_allocation=cost_allocation)

    # Assigning the OPEX data
    # fixed_cost_arr = df[[13, 16, 17]].sum(axis=1).to_numpy(dtype=float)
    # variable_cost_arr = df[14].to_numpy(dtype=float)
    # cost_per_volume_arr = df[15].to_numpy(dtype=float)

    # Reading the produced fluid for determining the prod_rate attributes of the OPEX dataclass
    # if cost_allocation == FluidType.OIL:
    #     prod_arr = df[20].to_numpy(dtype=float)
    #
    # elif cost_allocation == FluidType.GAS:
    #     prod_arr = df[22].to_numpy(dtype=float)
    #
    # else:
    #     prod_arr = df[21].to_numpy(dtype=float)
    #
    # opex = OPEX(start_year=start_year,
    #             end_year=end_year,
    #             fixed_cost=fixed_cost_arr,
    #             expense_year=years_arr,
    #             cost_allocation=cost_allocation,
    #             prod_rate=prod_arr,
    #             cost_per_volume=cost_per_volume_arr)

    return tangible, intangible, opex, asr

#
# if __name__ == "__main__":
#     # Choosing the Dataset and contract type
#     import timeit
#     start_time = timeit.default_timer()
#     dataset = 'small_gas'
#     contract = 'project'
#
#     # Returning the load_data function
#     psc = load_data(dataset_type=dataset, contract_type=contract)
#     print('Output of the load_data function \n', psc, '\n')
#     end_time = timeit.default_timer()
#     print("The Execution Time:", end_time - start_time)



"""
Calculate depreciation.
"""

import numpy as np


class DepreciationException(Exception):
    """Raise an exception error for a misuse of depreciation method"""

    pass


class UnitOfProductionException(Exception):
    """Exception to be raised for a misuse of unit of production method"""

    pass


def straight_line_depreciation_rate(
    cost: float, salvage_value: float, useful_life: int, depreciation_len: int = 0
) -> np.ndarray:
    """
    Calculate the straight-line depreciation charge for each period.

    Parameters
    ----------
    cost : float
        Cost of the asset.
    salvage_value : float
        Remaining value after depreciation.
    useful_life : int
        Duration for depreciation.
    depreciation_len : int, optional
        Length of the depreciation charge array beyond the useful life.
        The extended values will be set to zero. (default: 0)

    Returns
    -------
    depreciation_charge : numpy.ndarray
        The straight-line depreciation charge for each period.

    Notes
    -----
    The straight-line depreciation method allocates an equal amount of depreciation charge
    for each period over the useful life of an asset. The depreciation rate is calculated as:

    depreciation_rate = (cost - salvage_value) / useful_life

    The depreciation charge is then repeated for each period.

    If the `depreciation_len` is greater than the `useful_life`, the depreciation charge array
    is extended with zero values for the additional periods.
    """

    depreciation_rate = (cost - salvage_value) / useful_life
    depreciation_charge = np.repeat(depreciation_rate, useful_life)

    # Extend the depreciation charge array beyond useful life if needed
    if depreciation_len > useful_life:
        extension = np.zeros(int(depreciation_len - useful_life))
        depreciation_charge = np.concatenate((depreciation_charge, extension))

    return depreciation_charge


def straight_line_book_value(
    cost: float, salvage_value: float, useful_life: int, depreciation_len: int = 0
) -> np.ndarray:
    """
    Calculate the book value of an asset over time using the straight-line depreciation method.

    Parameters
    ----------
    cost : float
        Cost of the asset.
    salvage_value : float
        Remaining value after depreciation.
    useful_life : int
        Duration for depreciation.
    depreciation_len : int, optional
        Length of the book value array beyond the useful life.
        The extended values will be calculated based on the depreciation charge. (default: 0)

    Returns
    -------
    book_value : numpy.ndarray
        The book value of the asset for each period.

    Notes
    -----
    The straight-line depreciation method allocates an equal amount of depreciation charge
    for each period over the useful life of an asset. The depreciation charge is calculated
    using the `straight_line_depreciation_rate` function.

    The book value of the asset is calculated as the initial cost minus the cumulative sum of
    the depreciation charge for each period.

    If the `depreciation_len` is greater than the `useful_life`, the book value array is extended
    by calculating the remaining book values based on the depreciation charge.
    """

    # Calculate depreciation charge from function straight_line_depreciation_rate()
    depreciation_charge = straight_line_depreciation_rate(
        cost=cost,
        salvage_value=salvage_value,
        useful_life=useful_life,
        depreciation_len=depreciation_len,
    )

    # Calculate book value
    book_value = cost - np.cumsum(depreciation_charge)

    return book_value


def declining_balance_depreciation_rate(
    cost: float,
    salvage_value: float,
    useful_life: int,
    decline_factor: float = 1,
    depreciation_len: int = 0,
) -> np.ndarray:
    """
    Calculate the declining balance depreciation charge for each period.

    Parameters
    ----------
    cost : float
        Cost of the asset.
    salvage_value : float
        Remaining value after depreciation.
    useful_life : int
        Duration for depreciation.
    decline_factor : float, optional
        Depreciation factor. Usually the value is between 1 and 2.
        When the value is set as 2, it is called DDB (default: 1)
    depreciation_len : int, optional
        Length of the net book value array beyond its useful life.
        The extended values will be set as zero. (default: 0)

    Returns
    -------
    depreciation_charge : numpy.ndarray
        The depreciation charge for each period.
    """

    # Exponent value
    periods = np.arange(1, int(useful_life), 1, dtype=np.int_)

    # Depreciation factor
    depreciation_factor = decline_factor / useful_life

    # Depreciation charge
    depreciation_charge = (
        depreciation_factor * cost * np.power(1 - depreciation_factor, periods - 1)
    )

    # When depreciation charge reaches the salvage value
    if depreciation_charge.sum() > (cost - salvage_value):
        remaining_depreciation = cost - salvage_value - np.cumsum(depreciation_charge)
        remaining_depreciation = np.where(remaining_depreciation > 0, remaining_depreciation, 0)
        idx = np.argmin(remaining_depreciation)

        # Adjust the depreciation charge to take only the remainder when it reaches the salvage value
        depreciation_charge[idx] = (cost - salvage_value - np.cumsum(depreciation_charge)[idx - 1])
        depreciation_charge[idx + 1:] = 0

    # Add an element to array depreciation_charge, accounting the remaining unpaid cost
    depreciation_charge = np.concatenate(
        (
            depreciation_charge,
            cost - salvage_value - np.sum(depreciation_charge, keepdims=True),
        )
    )

    # Extend the depreciation charge array beyond useful life if project duration is longer than useful life
    if depreciation_len > useful_life:
        extension = np.zeros(int(depreciation_len - useful_life))
        depreciation_charge = np.concatenate((depreciation_charge, extension))

    return depreciation_charge


def declining_balance_book_value(
    cost: float,
    salvage_value: float,
    useful_life: int,
    decline_factor: int = 1,
    depreciation_len: int = 0,
) -> np.ndarray:
    """
    Calculate the net book value of a depreciated asset using the Decline Balance Method.

    Parameters
    ----------
    cost : float
        Cost of the asset.
    salvage_value : float
        Remaining value after depreciation.
    useful_life : int
        Duration for depreciation.
    decline_factor : int, optional
        Depreciation factor. Set to 1 for Decline Balance or 2 for Double Decline Balance.
        (default: 1)
    depreciation_len : int, optional
        Length of the net book value array beyond its useful life.
        The extended values will be set using the salvage value. (default: 0)

    Returns
    -------
    book_value : numpy.ndarray
        The book value of the depreciated asset.

    Notes
    -----
    The book value (V_t) at time t is calculated using the formula:

    .. math::

        V_t = (1 - d)^t \times V_0

    where:
    - V_t is the net book value at time t,
    - d is the depreciation factor per year,
    - t is the time in years, and
    - V_0 is the initial cost of the asset.
    """
    depreciation_charge = declining_balance_depreciation_rate(
        cost=cost,
        salvage_value=salvage_value,
        useful_life=useful_life,
        decline_factor=decline_factor,
        depreciation_len=depreciation_len,
    )
    book_value = cost - np.cumsum(depreciation_charge)

    return book_value


def psc_declining_balance_depreciation_rate(
    cost: float,
    useful_life: float,
    depreciation_factor: float = 0.5,
    depreciation_len: int = 0,
) -> np.ndarray:
    """
    Calculate the declining balance depreciation charges following
    psc declining balance method.

    Parameters:
    -----------
    cost : float
        Initial cost of the asset.

    useful_life : float
        Total useful life of the asset in periods.

    depreciation_factor : float, optional
        Depreciation factor for declining balance calculation (default is 0.5).

    depreciation_len : int, optional
        Length of the resulting depreciation charge array (default is 0).
        If specified and greater than the calculated length, the array will be extended with zeros.

    Returns:
    --------
    depreciation_charge: np.ndarray
        An array containing the calculated depreciation charges for each period.

    Notes:
    ------
    The value of depreciation_factor must fall within the following interval:
    0 <= depreciation_factor <= 1.

    A value that falls outside the prescribed interval is in contrast with the logical
    concept of psc declining balance.
    """

    if depreciation_factor > 1:
        raise DepreciationException(
            f"The value of depreciation_factor must fall within the following interval: "
            f"0 <= depreciation_factor <= 1 "
        )

    elif depreciation_factor < 0:
        raise DepreciationException(
            f"The value of depreciation_factor must fall within the following interval: "
            f"0 <= depreciation_factor <= 1 "
        )

    else:
        # Create an array of exponents
        periods = np.arange(1, int(useful_life), 1, dtype=np.int_)

        # Calculate depreciation charge
        depreciation_charge = (
            depreciation_factor * cost * np.power(1 - depreciation_factor, periods - 1)
        )

        # Modify depreciation charge; specify the last element as the remaining (unpaid) cost
        depreciation_charge = np.concatenate(
            (depreciation_charge, cost - np.sum(depreciation_charge, keepdims=True))
        )

        # Modify depreciation charge; accounting for project duration
        if depreciation_len > len(depreciation_charge):
            extension = np.zeros(int(depreciation_len - len(depreciation_charge)))
            depreciation_charge = np.concatenate((depreciation_charge, extension))

        return depreciation_charge


def psc_declining_balance_book_value(
    cost: float,
    useful_life: float,
    depreciation_factor: float = 0.5,
    depreciation_len: int = 0,
) -> np.ndarray:
    """
    Calculate the book value of an asset over time using declining balance depreciation method.

    Parameters:
    -----------
    cost : float
        Initial cost of the asset.

    useful_life : float
        Total useful life of the asset in periods.

    depreciation_factor : float, optional
        Depreciation factor for declining balance calculation (default is 0.5).

    depreciation_len : int, optional
        Length of the resulting book value array (default is 0).
        If specified and greater than the calculated length, the array will be extended.

    Returns:
    --------
    book_value: np.ndarray
        An array containing the calculated book values of the asset over time.

    Notes:
    ------
    This function calculates the book value of an asset over time using the declining balance
    depreciation method. The book value at each period is calculated as the difference between
    the initial cost and the cumulative sum of depreciation charges.

    If `depreciation_len` is provided and is greater than the calculated length,
    the resulting array is extended.
    """

    depreciation_charge = psc_declining_balance_depreciation_rate(
        cost=cost,
        useful_life=useful_life,
        depreciation_factor=depreciation_factor,
        depreciation_len=depreciation_len,
    )

    book_value = cost - np.cumsum(depreciation_charge)

    return book_value


def unit_of_production_rate(
    start_year_project: int,
    cost: float,
    prod: np.ndarray,
    prod_year: np.ndarray,
    salvage_value: float = 0.0,
    amortization_len: int = 0,
) -> np.ndarray:
    """
    Calculate amortization charge following unit of production method.

    Parameters
    ----------
    start_year_project : int
        The starting year of the project.
    cost : float
        Initial cost of the asset.
    prod : np.ndarray
        Hydrocarbon production starting from onstream year.
    prod_year : np.ndarray
        An array of years corresponding to the hydrocarbon production.
    salvage_value : float, optional
        The salvage value of the project at the end of its life (default is 0.0).
    amortization_len : int, optional
        The length of the amortization charge array (default is 0).
        If specified and greater than the calculated length, the array will be extended with zeros.

    Returns
    -------
    amortization_charge: no.ndarray
        An array representing the amortization charge following 2 * unit_of_production.

    Notes
    -----
    The function calculates the amortization charge using 2 * unit of production method.
    This method spreads the cost of the asset over its useful life based on the amount of production.
    """
    # Raise an exception if 'prod' is not given as a numpy.ndarray datatype
    if not isinstance(prod, np.ndarray):
        raise UnitOfProductionException(
            f"Parameter prod must be given as a numpy.ndarray datatype, not "
            f"{prod.__class__.__qualname__}."
        )

    # Raise an exception if 'prod_year' is not given as a numpy.ndarray datatype
    if not isinstance(prod_year, np.ndarray):
        raise UnitOfProductionException(
            f"Parameter prod_year must be given as a numpy.ndarray datatype, not "
            f"{prod_year.__class__.__qualname__}"
        )

    # Raise an exception for unequal length of arrays: prod and prod_year
    if len(prod) != len(prod_year):
        raise UnitOfProductionException(
            f"Unequal number of arrays: "
            f"prod: {len(prod)}, "
            f"prod_year: {len(prod_year)}."
        )

    # Raise an exception if salvage_value is larger than the associated cost
    if salvage_value > cost:
        raise UnitOfProductionException(
            f"Salvage value ({salvage_value}) is larger than the associated cost ({cost})."
        )

    # Raise an exception if prod_year is before the start year of the project
    if min(prod_year) < start_year_project:
        raise UnitOfProductionException(
            f"Production year ({min(prod_year)}) is before the start year "
            f"of the project ({start_year_project})."
        )

    # Raise an exception if prod_year is after the end year of the project
    if max(prod_year) > int(start_year_project + amortization_len - 1):
        raise UnitOfProductionException(
            f"Production year ({max(prod_year)}) is after the end year "
            f"of the project ({int(start_year_project + amortization_len - 1)})"
        )

    # Specify cumulative production
    cum_prod = np.sum(prod, dtype=np.float_)

    # Raise an exception for zero or negative value of cum_prod
    if cum_prod < 0:
        raise UnitOfProductionException(
            f"Inappropriate value of production data. "
            f"The sum of yearly_prod ({prod}) is {cum_prod}."
        )

    # Calculate amortization charge (1 * UOP); place it in order of project years
    amortization_charge = np.divide(prod, cum_prod, where=cum_prod != 0) * (cost - salvage_value)
    amortization_charge = np.bincount(prod_year - start_year_project, weights=amortization_charge)

    # Calculate amortization charge (2 * UOP); place it in order of project years
    amortization_charge = 2.0 * amortization_charge

    # Calculate the remaining amortization
    remaining_amortization = cost - salvage_value - np.cumsum(amortization_charge)
    remaining_amortization = np.where(remaining_amortization > 0, remaining_amortization, 0)
    idx = np.argmin(remaining_amortization)

    # Modify amortization charge, accounting for the remaining amortization after 2 * UOP
    amortization_charge[idx] = remaining_amortization[int(idx - 1)]
    amortization_charge[idx + 1:] = 0.0

    # Modify amortization charge, accounting for project duration
    if amortization_len > len(amortization_charge):
        extension = np.zeros(int(amortization_len - len(amortization_charge)))
        amortization_charge = np.concatenate((amortization_charge, extension))

    return amortization_charge


def unit_of_production_book_value(
    start_year_project: int,
    cost: float,
    prod: np.ndarray,
    prod_year: np.ndarray,
    salvage_value: float = 0.0,
    amortization_len: int = 0,
) -> np.ndarray:
    """
    Calculate the book value of an asset over time using unit of production method.

    Parameters
    ----------
    start_year_project : int
        The starting year of the project.
    cost : float
        Initial cost of the asset.
    prod : np.ndarray
        Hydrocarbon production starting from onstream year.
    prod_year : np.ndarray
        An array of years corresponding to the hydrocarbon production.
    salvage_value : float, optional
        The salvage value of the project at the end of its life (default is 0.0).
    amortization_len : int, optional
        The length of the amortization charge array (default is 0).
        If specified and greater than the calculated length, the array will be extended with zeros.

    Returns
    --------
    book_value: np.ndarray
        An array containing the calculated book values of the asset over time.
    """

    amortization_charge = unit_of_production_rate(
        start_year_project=start_year_project,
        cost=cost,
        prod=prod,
        prod_year=prod_year,
        salvage_value=salvage_value,
        amortization_len=amortization_len,
    )

    book_value = cost - np.cumsum(amortization_charge)

    return book_value



"""
This file containing the tools which utilized by API adapter.
"""
from datetime import datetime, date
from typing import Dict

from pydantic import BaseModel
import numpy as np

from pyscnomics.econ.costs import CapitalCost, Intangible, OPEX, ASR
from pyscnomics.dataset.sample import assign_lifting, read_fluid_type
from pyscnomics.econ.selection import TaxRegime, TaxType, FTPTaxRegime, GrossSplitRegime
from pyscnomics.tools.helper import (get_inflation_applied_converter,
                                     get_npv_mode_converter,
                                     get_discounting_mode_converter,
                                     get_depreciation_method_converter,
                                     get_other_revenue_converter,
                                     get_split_type_converter,
                                     get_optimization_target_converter,
                                     get_optimization_parameter_converter)


class SetupBM(BaseModel):
    """
    The BaseModel to validate the API setup input data.

    Parameters
    ----------
    start_date : str | int
        The start date of the project.
    end_date : str | int
        The end date of the project.
    oil_onstream_date: str | int | None
        The start date of oil production.
    gas_onstream_date: str | int | None
        The start date of gas production.
    """
    start_date: str | int = "01/01/2010"
    end_date: str | int = "31/12/2045"
    oil_onstream_date: str | int | None = "01/01/2023"
    gas_onstream_date: str | int | None = "01/01/2023"


class SummaryArgumentsBM(BaseModel):
    """
    BaseModel to validate the API summary arguments input data.

    Parameters
    ----------
    reference_year: int
        The reference year used to calculate the economic indicator.
    inflation_rate: float
        The inflation rate used to calculate the economic indicator and cost inflation.
    discount_rate: float
        The discount rate used to calculate the discounted economic indicator.
    npv_mode: str
        The mode of the NPV used in the NPV calculation. The available option are:
        [SKK Full Cycle Real Terms, SKK Full Cycle Nominal Terms,
        Full Cycle Real Terms, Full Cycle Nominal Terms, Point Forward]
    discounting_mode: str
        The discounting mode used in the NPV calculation. The available option are: [End Year, Mid Year]

    """
    reference_year: int = 2022
    inflation_rate: float | int | list = 0.1
    discount_rate: float | int = 0.1
    npv_mode: str = "SKK Full Cycle Nominal Terms"
    discounting_mode: str = "End Year"
    profitability_discounted: bool = False


class CostRecoveryBM(BaseModel):
    """
    The BaseModel to validate the API Cost Recovery input data.

    Parameters
    ----------
    oil_ftp_is_available: bool
        The oil ftp availability of the cost recovery contract.
    oil_ftp_is_shared: bool
        The oil shared ftp condition of the cost recovery contract.
    oil_ftp_portion: float
        The oil ftp portion.
    gas_ftp_is_available: bool
        The gas ftp availability of the cost recovery contract.
    gas_ftp_is_shared: bool
        The gas shared ftp condition of the cost recovery contract.
    gas_ftp_portion: float
        The gas ftp portion.
    tax_split_type: str
        The tax split type used in the contract. The available option are: [Conventional, RC Split, ICP Sliding Scale].
    condition_dict: dict
        The condition input for the tax split type except for the conventional split.
    indicator_rc_icp_sliding: list
        The indicator input for the tax split type except for the conventional split.
    oil_ctr_pretax_share: float
        The oil contractor pre-tax share.
    gas_ctr_pretax_share: float
        The gas contractor pre-tax share.
    oil_ic_rate: float
        The oil Investment Credit (IC) rate.
    gas_ic_rate: float
        The gas Investment Credit (IC) rate.
    ic_is_available: bool
        The Investment Credit (IC) availability.
    oil_cr_cap_rate: float
        The oil cost recovery cap rate
    gas_cr_cap_rate: float
        The gas cost recovery cap rate
    oil_dmo_volume_portion: float
        The oil Domestic Market Obligation (DMO) volume portion.
    oil_dmo_fee_portion: float
        The oil Domestic Market Obligation (DMO) fee portion.
    oil_dmo_holiday_duration: float
        The oil Domestic Market Obligation (DMO) Holiday duration.
    gas_dmo_volume_portion: float
        The gas Domestic Market Obligation (DMO) volume portion.
    gas_dmo_fee_portion: float
        The gas Domestic Market Obligation (DMO) fee portion.
    gas_dmo_holiday_duration: float
        The gas Domestic Market Obligation (DMO) Holiday duration.
    """
    oil_ftp_is_available: bool = True
    oil_ftp_is_shared: bool = True
    oil_ftp_portion: float | int = 0.2
    gas_ftp_is_available: bool = True
    gas_ftp_is_shared: bool = True
    gas_ftp_portion: float | int = 0.2
    tax_split_type: str = "Conventional"
    condition_dict: dict = {}
    indicator_rc_icp_sliding: list[float] = []
    oil_ctr_pretax_share: float | int = 0.34722220
    gas_ctr_pretax_share: float | int = 0.5208330
    oil_ic_rate: float | int = 0
    gas_ic_rate: float | int = 0
    ic_is_available: bool = False
    oil_cr_cap_rate: float | int = 1
    gas_cr_cap_rate: float | int = 1
    oil_dmo_volume_portion: float | int = 0.25
    oil_dmo_fee_portion: float | int = 0.25
    oil_dmo_holiday_duration: float | int = 60
    gas_dmo_volume_portion: float | int = 0.25
    gas_dmo_fee_portion: float | int = 1
    gas_dmo_holiday_duration: float | int = 60


class GrossSplitBM(BaseModel):
    """
    The BaseModel to validate the API Gross Split input data.

    Parameters
    ----------
    field_status: str
        The field status. The available options are:
        [POD I, POD II, POFD, No POD].
    field_loc: str
        The field location. The available options are:
        [Onshore, Offshore (0<h<=20), Offshore (20<h<=50),
        Offshore (50<h<=150), Offshore (150<h<=1000), Offshore (h>1000)].
    res_depth: str
        The reservoir depth of the field. The available options are:
        [<=2500, >2500]
    infra_avail: str
        The infrastructure availability of the field. The available options are:
        [Well Developed, New Frontier].
    res_type: str
        The reservoir type of the field. The available options are:
        [Conventional, Non Conventional]
    api_oil: str
        The oil API classification. The available options are:
        [<25, >=25].
    domestic_use: str
        The local content of the contract. The available options are:
        [<30, 30<=x<50, 50<=x<70, 70<=x<100]
    prod_stage: str
        The production stage of the contract. The available options are:
        [Primary, Secondary, Tertiary]
    co2_content: str
        The CO2 content of the field produced fluid. The available options are:
        [<5, 5<=x<10, 10<=x<20, 20<=x<40, 40<=x<60, x>=60]
    h2s_content: str
        The H2S content of the field produced fluid. The available options are:
        [<100, 100<=x<300, 300<=x<500, x>=500]
    base_split_ctr_oil: float
        The contractor base split for oil.
    base_split_ctr_gas: float
        The contractor base split for gas.
    split_ministry_disc: float
        The ministerial discretion split.
    oil_dmo_volume_portion: float
        The oil Domestic Market Obligation (DMO) volume portion.
    oil_dmo_fee_portion: float
        The oil Domestic Market Obligation (DMO) fee portion.
    oil_dmo_holiday_duration: float
        The oil Domestic Market Obligation (DMO) Holiday duration.
    gas_dmo_volume_portion: float
        The gas Domestic Market Obligation (DMO) volume portion.
    gas_dmo_fee_portion: float
        The gas Domestic Market Obligation (DMO) fee portion.
    gas_dmo_holiday_duration: float
        The gas Domestic Market Obligation (DMO) Holiday duration.

    """
    field_status: str = "No POD"
    field_loc: str = "Onshore"
    res_depth: str = "<=2500"
    infra_avail: str = "Well Developed"
    res_type: str = "Conventional"
    api_oil: str = "<25"
    domestic_use: str = "50<=x<70"
    prod_stage: str = "Secondary"
    co2_content: str = "<5"
    h2s_content: str = "<100"
    base_split_ctr_oil: float | int = 0.43
    base_split_ctr_gas: float | int = 0.48
    split_ministry_disc: float | int = 0.08
    oil_dmo_volume_portion: float | int = 0.25
    oil_dmo_fee_portion: float | int = 1.0
    oil_dmo_holiday_duration: float | int = 60
    gas_dmo_volume_portion: float | int = 1.0
    gas_dmo_fee_portion: float | int = 1.0
    gas_dmo_holiday_duration: float | int = 60


class ContractArgumentsBM(BaseModel):
    """
    The BaseModel to validate the Contract Arguments input data.

    Parameters
    ----------
    sulfur_revenue: str
        The treatment of the sulfur revenue. The options are:
        [Addition to Oil Revenue, Addition to Gas Revenue, Reduction to Oil OPEX, Reduction to Gas OPEX]
    electricity_revenue: str
        The treatment of the electricity revenue. The options are:
        [Addition to Oil Revenue, Addition to Gas Revenue, Reduction to Oil OPEX, Reduction to Gas OPEX]
    co2_revenue: str
        The treatment of the CO2 revenue. The options are:
        [Addition to Oil Revenue, Addition to Gas Revenue, Reduction to Oil OPEX, Reduction to Gas OPEX]
    is_dmo_end_weighted: bool
        The option treatment of the DMO.
    tax_regime: str
        The tax regime used in tax calculation. The options are:
        [nailed, prevailing, UU No.36 Tahun 2008, UU No.02 Tahun 2020, UU No.07 Tahun 2021]
    tax_rate: float
        The tax rate used for the tax payment.
    ftp_tax_regime: str
        The tax scheme used to calculate the tax payment. The options are:
        [PDJP No.20 Tahun 2017, Pre PDJP No.20 Tahun 2017, Direct]
    sunk_cost_reference_year: int
        The sunk cost reference year.
    depr_method: str
        The depreciation method used for calculating the depreciation. The options are:
        [PSC Declining Balance, Declining Balance, Unit Of Production, Straight Line]
    decline_factor: float
        The decline factor for the depreciation if the depreciation method is Declining Balance.
    vat_rate: list
        The Value Added Tax (VAT) rate used in the calculation.
    lbt_rate: list
        The Land and Building (LBT) rate used in the calculation.
    inflation_rate: list
        The inflation rate used in the calculation.
    future_rate: float
        The future rate used in the ASR calculation.
    inflation_rate_applied_to: str
        The option of where the inflation will be applied to. The options are:
        [CAPEX, OPEX, CAPEX AND OPEX]

    """
    sulfur_revenue: str = "Addition to Oil Revenue"
    electricity_revenue: str = "Addition to Oil Revenue"
    co2_revenue: str = "Addition to Oil Revenue"
    is_dmo_end_weighted: bool = False
    tax_regime: str = "nailed down"
    tax_rate: float | list | int | None = 0.424
    ftp_tax_regime: str = "Pre PDJP No.20 Tahun 2017"
    sunk_cost_reference_year: int | None = 2021
    depr_method: str = "PSC Declining Balance"
    decline_factor: float | int = 2
    vat_rate: list | float | int = 0.0
    lbt_rate: list | float | int = 0.0
    inflation_rate: list | float | int = 0.0
    future_rate: float = 0.02
    inflation_rate_applied_to: str = "CAPEX"
    post_uu_22_year2001: bool = True
    cum_production_split_offset: list | float | int | None = None
    amortization: bool = False
    regime: str = "PERMEN_ESDM_12_2020"
    sum_undepreciated_cost: bool = False


class ContractArgumentsTransitionBM(BaseModel):
    """
    The BaseModel to validate the Contract Arguments of Transition input data.

    Parameters
    ----------
    unrec_portion: float
        The unrec portion that will be transferred into second contract.
    """
    unrec_portion: float | int


class LiftingBM(BaseModel):
    """
    The BaseModel to validate the Lifting input data.

    Parameters
    ----------
    start_year: int
        The start year of the lifting.
    end_year: int
        The end year of the lifting.
    lifting_rate: list[float]
        The list containing the lifting rate.
    price: list[float]
        The list containing the price.
    prod_year: list[int]
        The list containing the production year of the corresponding lifting.
    fluid_type: str
        The type of the corresponding fluid.
    ghv: list[float] | None
        The list containing the Gross Heating Value (GHV) of the corresponding fluid.
    prod_rate: list[float] | None
        The list containing the production rate of the corresponding lifting.
    prod_rate_baseline: list[float] | list[int] | None
        The list containing the production rate baseline of the corresponding lifting.
    """
    start_year: int
    end_year: int
    lifting_rate: list[float] | list[int]
    price: list[float] | list[int]
    prod_year: list[int]
    fluid_type: str
    ghv: list[float] | list[int] | None = None
    prod_rate: list[float] | list[int] | None = None
    prod_rate_baseline: list[float] | list[int] | None = None


class TangibleBM(BaseModel):
    """
    The BaseModel to validate the Tangible input data.

    Parameters
    ----------
    start_year: int
        The start year of the project.
    end_year: int
        The end year of the project.
    cost: list[float]
        An list representing the cost of a tangible asset.
    expense_year: list[int]
        An list representing the expense year of a tangible asset.
    cost_allocation: list[str]
        A list representing the cost allocation of a tangible asset.
    description: list[str]
        A list of string description regarding the associated tangible cost.
    vat_portion: list[float]
        The portion of 'cost' that is subject to VAT.
        Must be a list of length equals to the length of 'cost' list.
    vat_discount: list[float]
        The VAT discount to apply.
    lbt_portion: list[float]
        The portion of 'cost' that is subject to LBT.
        Must be a list of length equals to the length of 'cost' list.
    lbt_discount: list[float]
        The LBT discount to apply.
    pis_year: list[int]
        The list representing the PIS year of a tangible asset.
    salvage_value: list[float]
        The list representing the salvage value of a tangible asset.
    useful_life: list[int]
        The list representing the useful life of a tangible asset.
    depreciation_factor: list[float]
        The value of depreciation factor to be used in PSC DB depreciation method.
        Default value is 0.5 for the entire project duration.
    is_ic_applied: list[bool]
        The condition of will the IC applied or not.
    """

    start_year: int
    end_year: int
    cost: list[float] | list[int]
    expense_year: list[int]
    cost_allocation: list[str]
    description: list[str]
    vat_portion: list[float] | list[int]
    vat_discount: list[float] | list[int]
    lbt_portion: list[float] | list[int]
    lbt_discount: list[float] | list[int]
    pis_year: list[int]
    salvage_value: list[float] | list[int]
    useful_life: list[int]
    depreciation_factor: list[float] | list[int]
    is_ic_applied: list[bool]


class IntangibleBM(BaseModel):
    """
    The BaseModel to validate the Intangible input data.

    Parameters
    ----------
    start_year: int
        The start year of the project.
    end_year: int
        The end year of the project.
    cost: list[float]
        An list representing the cost of an intangible asset.
    expense_year: list[int]
        An list representing the expense year of an intangible asset.
    cost_allocation: list[str]
        A list representing the cost allocation of an intangible asset.
    description: list[str]
        A list of string description regarding the associated intangible cost.
    vat_portion: list[float]
        The portion of 'cost' that is subject to VAT.
        Must be a list of length equals to the length of 'cost' list.
    vat_discount: list[float]
        The VAT discount to apply.
    lbt_portion: list[float]
        The portion of 'cost' that is subject to LBT.
        Must be a list of length equals to the length of 'cost' list.
    lbt_discount: list[float]
        The LBT discount to apply.
    """
    start_year: int
    end_year: int
    cost: list[float] | list[int]
    expense_year: list[int]
    cost_allocation: list[str]
    description: list[str]
    vat_portion: list[float] | list[int]
    vat_discount: list[float] | list[int]
    lbt_portion: list[float] | list[int]
    lbt_discount: list[float] | list[int]


class OpexBM(BaseModel):
    """
    The BaseModel to validate the Opex input data.

    Parameters
    ----------
    start_year: int
        The start year of the project.
    end_year: int
        The end year of the project.
    expense_year: list[int]
        An list representing the expense year of an opex asset.
    cost_allocation: list[str]
        A list representing the cost allocation of an opex asset.
    description: list[str]
        A list of string description regarding the associated opex cost.
    vat_portion: list[float]
        The portion of 'cost' that is subject to VAT.
        Must be a list of length equals to the length of 'cost' list.
    vat_discount: list[float]
        The VAT discount to apply.
    lbt_portion: list[float]
        The portion of 'cost' that is subject to LBT.
        Must be a list of length equals to the length of 'cost' list.
    lbt_discount: list[float]
        The LBT discount to apply.
    fixed_cost : list
        An list representing the fixed cost of an OPEX asset.
    prod_rate: list
        The production rate of a particular fluid type.
    cost_per_volume: list
        Cost associated with production of a particular fluid type.
    """
    start_year: int
    end_year: int
    expense_year: list[int]
    cost_allocation: list[str]
    description: list[str]
    fixed_cost: list[float] | list[int]
    prod_rate: list[float] | list[int]
    cost_per_volume: list[float] | list[int]
    vat_portion: list[float] | list[int]
    vat_discount: list[float] | list[int]
    lbt_portion: list[float] | list[int]
    lbt_discount: list[float] | list[int]


class AsrBM(BaseModel):
    """
    The BaseModel to validate the ASR input data.

    Parameters
    ----------
    start_year: int
        The start year of the project.
    end_year: int
        The end year of the project.
    cost: list[float]
        An list representing the cost of an ASR asset.
    expense_year: list[int]
        An list representing the expense year of an ASR asset.
    cost_allocation: list[str]
        A list representing the cost allocation of an ASR asset.
    description: list[str]
        A list of string description regarding the associated intangible cost.
    vat_portion: list[float]
        The portion of 'cost' that is subject to VAT.
        Must be a list of length equals to the length of 'cost' list.
    vat_discount: list[float]
        The VAT discount to apply.
    lbt_portion: list[float]
        The portion of 'cost' that is subject to LBT.
        Must be a list of length equals to the length of 'cost' list.
    lbt_discount: list[float]
        The LBT discount to apply.
    """
    start_year: int
    end_year: int
    cost: list[float] | list[int]
    expense_year: list[int]
    cost_allocation: list[str]
    description: list[str]
    vat_portion: list[float] | list[int]
    vat_discount: list[float] | list[int]
    lbt_portion: list[float] | list[int]
    lbt_discount: list[float] | list[int]


class OptimizationDictBM(BaseModel):
    """
    The BaseModel to validate the Optimization Dictionary input data.

    Parameters
    ----------
    parameter: list[str]
        The list of string containing the parameter which will be obtained.
    min: list[float]
        The list of float containing the minimum boundary for the optimization variable.
    max: list[float]
        The list of float containing the maximum boundary for the optimization variable.

    """
    parameter: list[str]
    min: list[float] | list[int]
    max: list[float] | list[int]


class OptimizationBM(BaseModel):
    """
    The BaseModel to validate the Optimization input data.

    Parameters
    ----------
    dict_optimization: OptimizationDictBM
        The dictionary containing the optimization parameter
    target_optimization: float
        The value of the optimization target.
    target_parameter: str
        The targeted optimization parameter.
    """
    dict_optimization: OptimizationDictBM
    target_optimization: float | int
    target_parameter: str


class SensitivityBM(BaseModel):
    """
    The BaseModel to validate the Sensitivity input data.

    Parameters
    ----------
    min: float
        The minimum value for the sensitivity
    max: float
        The maximum value for the sensitivity
    """
    min: float | int
    max: float | int


class UncertaintyBM(BaseModel):
    """
    The BaseModel to validate the uncertainty input data.

    Parameters
    ----------
    number_of_simulation: int
        The number of the simulation.
    min: list[float]
        The minimum value for uncertainty model
    max: list[float]
        The maximum value for uncertainty model
    std_dev: list[float]
        The standard deviation for the uncertainty model
    """
    number_of_simulation: int
    min: list[float] | list[int]
    max: list[float] | list[int]
    std_dev: list[float] | list[int]


class Data(BaseModel):
    """
    The BaseModel to validate the Data input.

    Parameters
    ----------
    setup: SetupBM
        The setup input in form of SetupBM BaseModel.
    summary_arguments: SummaryArgumentsBM
        The summary arguments in form of SummaryArgumentsBM BaseModel.
    contract_arguments: ContractArgumentsBM
        The contract arguments in form of ContractArgumentsBM BaseModel.
    lifting: Dict[str, LiftingBM]
        The lifting input in form of LiftingBM BaseModel.
    tangible: Dict[str, TangibleBM]
        The tangible input in form of TangibleBM BaseModel.
    intangible: Dict[str, IntangibleBM]
        The intangible input in form of IntangibleBM BaseModel.
    opex: Dict[str, OpexBM]
        The opex input in form of OpexBM BaseModel.
    asr: Dict[str, AsrBM]
        The asr input in form of AsrBM BaseModel.
    optimization_arguments: OptimizationBM = None
        The optimization arguments in form of OptimizationBM BaseModel.
    sensitivity_arguments: SensitivityBM = None
        The sensitivity arguments in form of SensitivityBM BaseModel.
    uncertainty_arguments: UncertaintyBM = None
        The uncertainty arguments in form of UncertaintyBM BaseModel.
    costrecovery: CostRecoveryBM = None
        The cost recovery object in form of CostRecoveryBM BaseModel.
    grosssplit: GrossSplitBM = None
        The gross split object in form of GrossSplit BaseModel.
    result: dict = None
        The result of the running contract.
    """
    setup: SetupBM
    summary_arguments: SummaryArgumentsBM
    contract_arguments: ContractArgumentsBM
    lifting: Dict[str, LiftingBM]
    tangible: Dict[str, TangibleBM]
    intangible: Dict[str, IntangibleBM]
    opex: Dict[str, OpexBM]
    asr: Dict[str, AsrBM]
    optimization_arguments: OptimizationBM = None
    sensitivity_arguments: SensitivityBM = None
    uncertainty_arguments: UncertaintyBM = None
    costrecovery: CostRecoveryBM = None
    grosssplit: GrossSplitBM = None
    result: dict = None


class TransitionBM(BaseModel):
    """
    The BaseModel to validate the contract in the Transition input data.

    Parameters
    ----------
    setup: SetupBM
        The setup input in form of SetupBM BaseModel.
    contract_arguments: ContractArgumentsBM
        The contract arguments in form of ContractArgumentsBM BaseModel.
    lifting: Dict[str, LiftingBM]
        The lifting input in form of LiftingBM BaseModel.
    tangible: Dict[str, TangibleBM]
        The tangible input in form of TangibleBM BaseModel.
    intangible: Dict[str, IntangibleBM]
        The intangible input in form of IntangibleBM BaseModel.
    opex: Dict[str, OpexBM]
        The opex input in form of OpexBM BaseModel.
    asr: Dict[str, AsrBM]
        The asr input in form of AsrBM BaseModel.
    costrecovery: CostRecoveryBM = None
        The cost recovery object in form of CostRecoveryBM BaseModel.
    grosssplit: GrossSplitBM = None
        The gross split object in form of GrossSplit BaseModel.
    """
    setup: SetupBM
    contract_arguments: ContractArgumentsBM
    lifting: Dict[str, LiftingBM]
    tangible: Dict[str, TangibleBM]
    intangible: Dict[str, IntangibleBM]
    opex: Dict[str, OpexBM]
    asr: Dict[str, AsrBM]
    costrecovery: CostRecoveryBM | None = None
    grosssplit: GrossSplitBM | None = None


class DataTransition(BaseModel):
    """
    The BaseModel to validate the Transition Data input.

    Parameters
    ----------
    contract_1: TransitionBM
        The first contract object in form of TransitionBM
    contract_2: TransitionBM
        The second contract object in form of TransitionBM
    contract_arguments: ContractArgumentsTransitionBM
        The contract arguments in form of ContractArgumentsBM BaseModel.
    summary_arguments: SummaryArgumentsBM
        The summary arguments in form of SummaryArgumentsBM BaseModel.
    result: dict = None
        The result of the running contract.
    """
    contract_1: TransitionBM
    contract_2: TransitionBM
    contract_arguments: ContractArgumentsTransitionBM
    summary_arguments: SummaryArgumentsBM
    result: dict = None
    optimization_arguments: OptimizationBM = None


def convert_str_to_date(str_object: str | int) -> date | None:
    """
    The function to convert string or integer unix timestamp format object into dateformat
    Parameters
    ----------
    str_object: str | int
        The object which will be converted into date format

    Returns
    -------
    date

    """
    if str_object is None:
        return None
    else:
        if isinstance(str_object, str):
            return datetime.strptime(str_object, '%d/%m/%Y').date()
        elif isinstance(str_object, int):
            value = datetime.fromtimestamp(str_object)
            date_value = value.strftime('%d/%m/%Y')
            return datetime.strptime(date_value, '%d/%m/%Y').date()


def convert_list_to_array_float(data_list: list) -> np.ndarray:
    """
    The function to convert list into array of float.

    Parameters
    ----------
    data_list: list
        The list which will be converted into numpy array.

    Returns
    -------
    array
        np.ndarray
    """
    return np.array(data_list, dtype=float)


def convert_list_to_array_float_or_array(data_input: list | float | int | str) -> np.ndarray | float:
    """
    The function to convert list or float into float or array of float.

    Parameters
    ----------
    data_input: list | float
        The list or float which will be converted

    Returns
    -------
    output
        float | np.ndarray
    """
    if isinstance(data_input, float):
        return data_input

    elif isinstance(data_input, int) or isinstance(data_input, str):
        return (
            float(data_input)
            if isinstance(data_input, int)
               or (isinstance(data_input, str) and data_input.strip() != "")
            else 0
        )
    else:
        return np.array(data_input, dtype=float)


def convert_list_to_array_float_or_array_or_none(data_list: list | float | int | str | None
                                                 ) -> np.ndarray | float | None:
    """
    Function to convert list into array of float, None or array.

    Parameters
    ----------
    data_list
        The list which will be converted.

    Returns
    -------
    out
        np.ndarray | float | None

    """
    if isinstance(data_list, float):
        return data_list
    elif isinstance(data_list, int) or isinstance(data_list, str):
        return (
            float(data_list)
            if isinstance(data_list, int)
               or (isinstance(data_list, str) and data_list.strip() != "")
            else None
        )
    elif data_list is None:
        return None
    else:
        return np.array(data_list, dtype=float)


def convert_list_to_array_int(data_list: list) -> np.ndarray:
    """
    The function to convert list into array of integer.

    Parameters
    ----------
    data_list: list
        The list which will be converted into numpy array of integer.

    Returns
    -------
    out
        np. array

    """
    return np.array(data_list, dtype=int)


def convert_dict_to_lifting(data_raw: dict) -> tuple:
    """
    The function to convert dictionary into tuple of Lifting dataclass.

    Parameters
    ----------
    data_raw: dict
        The dictionary which will be converted into tuple of Lifting


    Returns
    -------
    out:
        tuple[Lifting]
    """
    return assign_lifting(data_raw=data_raw)


def convert_dict_to_capital(data_raw: dict) -> tuple:
    """
    The function to convert a dictionary into tuple of Tangible dataclass.

    Parameters
    ----------
    data_raw
        The dictionary which will be converted into tuple of Tangible

    Returns
    -------
    out:
        tuple[Tangible]
    """
    tangible_list = [
        CapitalCost(
            start_year=data_raw[key]['start_year'],
            end_year=data_raw[key]['end_year'],
            cost=np.array(data_raw[key]['cost']),
            expense_year=np.array(data_raw[key]['expense_year'], dtype=int),
            cost_allocation=read_fluid_type(fluid=data_raw[key]['cost_allocation']),
            description=data_raw[key]['description'],
            vat_portion=np.array(data_raw[key]['vat_portion']),
            vat_discount=np.array(data_raw[key]['vat_discount']),
            lbt_portion=np.array(data_raw[key]['lbt_portion']),
            lbt_discount=np.array(data_raw[key]['lbt_discount']),
            pis_year=np.array(data_raw[key]['pis_year']),
            salvage_value=np.array(data_raw[key]['salvage_value']),
            useful_life=np.array(data_raw[key]['useful_life']),
            depreciation_factor=np.array(data_raw[key]['depreciation_factor']),
            is_ic_applied=data_raw[key]['is_ic_applied'],
        )
        for key in data_raw.keys()
    ]

    return tuple(tangible_list)


def convert_dict_to_intangible(data_raw: dict) -> tuple:
    """
    The function to convert dictionary into tuple of Intangible dataclass.

    Parameters
    ----------
    data_raw: dict
        The dictionary which will be converted into tuple of Intangible

    Returns
    -------
    out:
        tuple[Intangible]
    """
    intangible_list = [
        Intangible(
            start_year=data_raw[key]['start_year'],
            end_year=data_raw[key]['end_year'],
            cost=np.array(data_raw[key]['cost'], dtype=float),
            expense_year=np.array(data_raw[key]['expense_year'], dtype=int),
            cost_allocation=read_fluid_type(fluid=data_raw[key]['cost_allocation']),
            description=data_raw[key]['description'],
            vat_portion=np.array(data_raw[key]['vat_portion'], dtype=float),
            vat_discount=np.array(data_raw[key]['vat_discount'], dtype=float),
            lbt_portion=np.array(data_raw[key]['lbt_portion'], dtype=float),
            lbt_discount=np.array(data_raw[key]['lbt_discount'], dtype=float),
        )
        for key in data_raw.keys()
    ]

    return tuple(intangible_list)


def convert_dict_to_opex(data_raw: dict) -> tuple:
    """
    The function to convert dictionary into tuple of OPEX dataclass.

    Parameters
    ----------
    data_raw: dict
        The dictionary which will be converted into tuple of OPEX

    Returns
    -------
    out:
        tuple[OPEX]
    """
    opex_list = [
        OPEX(
            start_year=data_raw[key]['start_year'],
            end_year=data_raw[key]['end_year'],
            expense_year=np.array(data_raw[key]['expense_year'], dtype=int),
            cost_allocation=read_fluid_type(fluid=data_raw[key]['cost_allocation']),
            description=data_raw[key]['description'],
            vat_portion=np.array(data_raw[key]['vat_portion'], dtype=float),
            vat_discount=np.array(data_raw[key]['vat_discount'], dtype=float),
            lbt_portion=np.array(data_raw[key]['lbt_portion'], dtype=float),
            lbt_discount=np.array(data_raw[key]['lbt_discount'], dtype=float),
            fixed_cost=np.array(data_raw[key]['fixed_cost'], dtype=float),
            prod_rate=np.array(data_raw[key]['prod_rate'], dtype=float),
            cost_per_volume=np.array(data_raw[key]['cost_per_volume'], dtype=float),
        )
        for key in data_raw.keys()
    ]

    return tuple(opex_list)


def convert_dict_to_asr(data_raw: dict) -> tuple:
    """
    The function to convert dictionary into tuple of ASR dataclass.

    Parameters
    ----------
    data_raw: dict
        The dictionary which will be converted into tuple of ASR

    Returns
    -------
    out:
        tuple[ASR]
    """
    asr_list = [
        ASR(
            start_year=data_raw[key]['start_year'],
            end_year=data_raw[key]['end_year'],
            cost=np.array(data_raw[key]['cost'], dtype=float),
            expense_year=np.array(data_raw[key]['expense_year'], dtype=int),
            cost_allocation=read_fluid_type(fluid=data_raw[key]['cost_allocation']),
            description=data_raw[key]['description'],
            vat_portion=np.array(data_raw[key]['vat_portion'], dtype=float),
            vat_discount=np.array(data_raw[key]['vat_discount'], dtype=float),
            lbt_portion=np.array(data_raw[key]['lbt_portion'], dtype=float),
            lbt_discount=np.array(data_raw[key]['lbt_discount'], dtype=float),
        )
        for key in data_raw.keys()
    ]

    return tuple(asr_list)


def convert_str_to_taxsplit(str_object: str):
    """
    The function to converts a string representing a tax split type to its corresponding
    enum value from the TaxSplitTypeCR class.

    Parameters
    ----------
    str_object: str
        The string representation of the tax split type.

    Returns
    -------
    Union: TaxSplitTypeCR, None
        The corresponding enum value if the target matches one of the predefined split types.
        Returns None if no match is found.

    Example
    -------
    >>> result = get_split_type_converter("RC Split")
    >>> print(result)
    TaxSplitTypeCR.R2C
    """
    return get_split_type_converter(target=str_object)


def convert_str_to_npvmode(str_object: str):
    """
    The function to converts a string representing an NPV mode to its corresponding
    enum value from the NPVSelection class.

    Parameters
    ----------
    str_object: str
        The string representation of the NPV mode.

    Returns
    -------
    Union[NPVSelection, None]
        The corresponding enum value if the target matches one of the predefined NPV modes.
        Returns None if no match is found.

    Example
    -------
    >>> result = get_npv_mode_converter("SKK Full Cycle Real Terms")
    >>> print(result)
    NPVSelection.NPV_SKK_REAL_TERMS
    """
    return get_npv_mode_converter(target=str_object)


def convert_str_to_discountingmode(str_object: str):
    """
    The function to converts a string representing a discounting mode to its corresponding
    enum value from the DiscountingMode class.

    Parameters
    ----------
    str_object: str
        The string representation of the discounting mode.

    Returns
    -------
        Union[DiscountingMode, None]
        The corresponding enum value if the target matches one of the predefined discounting modes.
        Returns None if no match is found.

    Example
    -------
    >>> result = get_discounting_mode_converter("End Year")
    >>> print(result)
    DiscountingMode.END_YEAR

    """
    return get_discounting_mode_converter(target=str_object)


def convert_str_to_otherrevenue(str_object: str):
    """
    The function to convert a string representation of other revenue types to the corresponding enum value.

    Parameters
    ----------
    str_object: str
        The string representation of the other revenue type.

    Returns
    -------
    OtherRevenue or None
        The enum value corresponding to the provided string representation of other revenue.
        Returns None if the string does not match any known other revenue types.

    Notes
    -----
    - The function uses a mapping of string representations to OtherRevenue enum values.
    - If the target string matches a known other revenue type, the corresponding enum value is returned.
    - If the target string does not match any known types, the function returns None.
    """
    return get_other_revenue_converter(target=str_object)


def convert_str_to_taxregime(str_object: str) -> TaxRegime | None:
    """
    The function to convert string into tax regime selection.

    Parameters
    ----------
    str_object: str
        The string object which will be converted

    Returns
    -------
    out:
        TaxRegime | None
    """
    dict_tax_regime = {i.value: i for i in TaxRegime}

    for key in dict_tax_regime.keys():
        if str_object == key:
            return dict_tax_regime[key]
        else:
            return None


def convert_str_to_ftptaxregime(str_object: str):
    """
    The functon to converts a string representing a tax regime to its corresponding
    enum value from the FTPTaxRegime class.

    Parameters
    ----------
    str_object: str
        The string representation of the tax regime.


    Returns
    -------
    Union[FTPTaxRegime, None]
        The corresponding enum value if the target matches one of the predefined tax regimes.
        Returns None if no match is found.

    """
    attrs = {
        "PDJP No.20 Tahun 2017": FTPTaxRegime.PDJP_20_2017,
        "Pre PDJP No.20 Tahun 2017": FTPTaxRegime.PRE_PDJP_20_2017,
        "Direct Mode": FTPTaxRegime.DIRECT_MODE
    }

    for key in attrs.keys():
        if str_object == key:
            return attrs[key]


def convert_str_to_depremethod(str_object: str):
    """
    The function to converts a string representing a depreciation method to its corresponding
    enum value from the DeprMethod class.

    Parameters
    ----------
    str_object: str
        The string representation of the depreciation method.

    Returns
    -------
    Union[DeprMethod, None]
        The corresponding enum value if the target matches one of the predefined depreciation methods.
        Returns None if no match is found.

    Example
    -------
    >>> result = get_depreciation_method_converter("PSC Declining Balance")
    >>> print(result)
    DeprMethod.PSC_DB

    """
    return get_depreciation_method_converter(target=str_object)


def convert_str_to_taxtype(str_object: str):
    """
    The function to convert a string object into tax type selection.

    Parameters
    ----------
    str_object: str
        The string representation of tax type.

    Returns
    -------
    out:
        TaxType
    """
    dict_tax_regime = {i.value: i for i in TaxType}

    for key in dict_tax_regime.keys():
        if str_object == key:
            return dict_tax_regime[key]
        else:
            return None


def convert_str_to_inflationappliedto(str_object: str):
    """
    The function to converts a string representing the application of inflation to its
    corresponding enum value from the InflationAppliedTo class.

    Parameters
    ----------
    str_object: str
        The string representation of the inflation application.

    Returns
    -------
    Union[InflationAppliedTo, None]
        The corresponding enum value if the target matches one of the predefined options.
        Returns None if no match is found.

    Example:
    >>> result = get_inflation_applied_converter("OPEX")
    >>> print(result)
    InflationAppliedTo.OPEX

    """
    return get_inflation_applied_converter(target=str_object)


def convert_grosssplitregime_to_enum(target: str) -> GrossSplitRegime:
    """
    Converts a string representing the Gross Split Regime to its
    corresponding enum value from the GrossSplitRegime enum class.

    Parameters
    ----------
    target: str
        The string representation of the gross split regime.

    Returns
    -------

    """
    attrs = {
        "PERMEN_ESDM_8_2017": GrossSplitRegime.PERMEN_ESDM_8_2017,
        "PERMEN_ESDM_52_2017": GrossSplitRegime.PERMEN_ESDM_52_2017,
        "PERMEN_ESDM_20_2019": GrossSplitRegime.PERMEN_ESDM_20_2019,
        "PERMEN_ESDM_12_2020": GrossSplitRegime.PERMEN_ESDM_12_2020,
    }

    for key in attrs.keys():
        if target == key:
            return attrs[key]


def convert_summary_to_dict(dict_object: dict):
    """
    The function to convert the summary into skk executive summary format.
    Parameters
    ----------
    dict_object: dict
        The dictionary representation of the summary.

    Returns
    -------
    out: dict
        The dictionary of skk executive summary.

    """
    summary_skk_format = {
        'lifting_oil': dict_object['lifting_oil'],
        'oil_wap': dict_object['oil_wap'],
        'lifting_gas': dict_object['lifting_gas'],
        'gas_wap': dict_object['gas_wap'],
        'gross_revenue': dict_object['gross_revenue'],
        'ctr_gross_share': dict_object['ctr_gross_share'],
        'sunk_cost': dict_object['sunk_cost'],
        'investment': dict_object['investment'],
        'tangible': dict_object['tangible'],
        'intangible': dict_object['intangible'],
        'opex_and_asr': dict_object['opex_and_asr'],
        'opex': dict_object['opex'],
        'asr': dict_object['asr'],
        'cost_recovery/deductible_cost': dict_object['cost_recovery / deductible_cost'],
        'cost_recovery_over_gross_rev': dict_object['cost_recovery_over_gross_rev'],
        'unrec_cost': dict_object['unrec_cost'],
        'unrec_over_gross_rev': dict_object['unrec_over_gross_rev'],
        'ctr_net_share': dict_object['ctr_net_share'],
        'ctr_net_share_over_gross_share': dict_object['ctr_net_share_over_gross_share'],
        'ctr_net_cashflow': dict_object['ctr_net_cashflow'],
        'ctr_net_cashflow_over_gross_rev': dict_object['ctr_net_cashflow_over_gross_rev'],
        'ctr_npv': dict_object['ctr_npv'],
        'ctr_irr': dict_object['ctr_irr'],
        'ctr_pot': dict_object['ctr_pot'],
        'ctr_pv_ratio': dict_object['ctr_pv_ratio'],
        'ctr_pi': dict_object['ctr_pi'],
        'gov_gross_share': dict_object['gov_gross_share'],
        'gov_ftp_share': dict_object['gov_ftp_share'],
        'gov_ddmo': dict_object['gov_ddmo'],
        'gov_tax_income': dict_object['gov_tax_income'],
        'gov_take': dict_object['gov_take'],
        'gov_take_over_gross_rev': dict_object['gov_take_over_gross_rev'],
        'gov_take_npv': dict_object['gov_take_npv'],
        'undepreciated_asset_oil': dict_object['undepreciated_asset_oil'],
        'undepreciated_asset_gas': dict_object['undepreciated_asset_gas'],
        'undepreciated_asset_total': dict_object['undepreciated_asset_total']

    }
    return summary_skk_format


def convert_str_to_optimization_targetparameter(str_object: str):
    """
    The function to converts a string representation of an optimization target to the corresponding
    OptimizationTarget enum value.

    Parameters
    ----------
    str_object: str
        The string representation of the optimization target.

    Returns
    -------
    Union[OptimizationTarget, None]
        The OptimizationTarget enum value if the target is valid, or None if the target is not found.

    Example
    -------
    >>> get_optimization_target_converter("NPV")
    OptimizationTarget.NPV
    >>> get_optimization_target_converter("IRR")
    OptimizationTarget.IRR
    >>> get_optimization_target_converter("PI")
    OptimizationTarget.PI
    >>> get_optimization_target_converter("InvalidTarget")
    None
    """
    return get_optimization_target_converter(target=str_object)


def convert_str_to_optimization_parameters(str_object: str):
    """
    The function to converts a string representation of an optimization parameter to the corresponding
    OptimizationParameter enum value.

    Parameters
    ----------
    str_object: str
        The string representation of the optimization parameter.

    Returns
    -------
    Union[OptimizationParameter, None]
        The OptimizationParameter enum value if the target is valid,
        or None if the target is not found.

    Example
    -------
    >>> get_optimization_parameter_converter("Oil Contractor Pre Tax")
    OptimizationParameter.OIL_CTR_PRETAX
    >>> get_optimization_parameter_converter("Gas FTP Portion")
    OptimizationParameter.GAS_FTP_PORTION
    >>> get_optimization_parameter_converter("VAT Rate")
    OptimizationParameter.VAT_RATE
    >>> get_optimization_parameter_converter("InvalidParameter")
    None

    """
    return get_optimization_parameter_converter(target=str_object)


def convert_to_float(target=int):
    """
    Function to convert integer into float.

    Parameters
    ----------
    target: int
        The target that will be converted.

    Returns
    -------
    float
    """
    return float(target)
